from fetchfox.constants.algorand.basics import ALGORAND
from fetchfox.constants.cardano.basics import CARDANO
from fetchfox.constants.ethereum.basics import ETHEREUM
from fetchfox.constants.polygon.basics import POLYGON


BLOCKCHAINS = [
    ALGORAND,
    CARDANO,
    ETHEREUM,
    POLYGON,
]
