from fetchfox.constants.algorand.marketplaces import ALGOXNFT_COM, RANDGALLERY_COM, SHUFL_APP
from fetchfox.constants.cardano.marketplaces import JPG_STORE
from fetchfox.constants.evm.marketplaces import OPENSEA_IO, RARIBLE_COM


UNKNOWN = "unknown"


MARKETPLACES = [
    ALGOXNFT_COM,
    RANDGALLERY_COM,
    SHUFL_APP,
    JPG_STORE,
    OPENSEA_IO,
    RARIBLE_COM,
]
